/*
 * This file is part of GtkGraphView.
 *
* Foobar is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* GtkGraphView is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with GtkGraphView. If not, see <https://www.gnu.org/licenses/>.
*/

#include <gtk-graph-view.h>
#include <epoxy/gl.h>
#include <shimatta-opengl-program.h>
#include <shimatta-opengl-colored-triangle.h>

#ifndef SHADER_DIRECTORY
#define SHADER_DIRECTORY "/usr/share/gtk-graph-view/shaders"
#endif

struct _GtkGraphView {
	GtkBox super;
};

typedef struct {
	/**
	 * @brief The gl area displaying the graph view.
	 * @note This variable is not referenced, because the gl_area is tightly bound to the GraphView widget itself.
	 */
	GtkWidget *gl_area;
	GdkRGBA background_color;
	ShimattaOpenglProgram *passthrough_prog;
	ShimattaOpenglGraphics *test;

} GtkGraphViewPrivate;

G_DEFINE_TYPE_WITH_PRIVATE(GtkGraphView, gtk_graph_view, GTK_TYPE_BOX)

enum {
	PROP_BG_COLOR = 1,
	N_PROPERTIES
};

/**
 * @brief Make the gl area the current context and set the clear color
 * @param[in] gl_area openGL area
 * @param[in] rgba Clear color
 */
static void gl_area_set_clear_color(GtkGLArea *gl_area, const GdkRGBA *rgba)
{
	g_return_if_fail(gl_area);
	g_return_if_fail(rgba != NULL);
	gtk_gl_area_make_current(gl_area);
	glClearColor(rgba->red, rgba->green, rgba->blue, rgba->alpha);
}

static gboolean gl_render(GtkGLArea *gl_area, GdkGLContext *context, GtkGraphView *view)
{
	GtkGraphViewPrivate *priv;

	priv = gtk_graph_view_get_instance_private(view);

	gl_area_set_clear_color(gl_area, &priv->background_color);
	glClear(GL_COLOR_BUFFER_BIT);

	shimatta_opengl_graphics_draw(priv->test);

	glFlush();

	return TRUE;
}

static void gtk_graph_view_get_property(GObject *obj, guint property_id, GValue *value, GParamSpec *pspec)
{
}

static void gtk_graph_view_set_property(GObject *obj, guint property_id, const GValue *value, GParamSpec *pspec)
{
	gpointer data_ptr;
	GtkGraphViewPrivate *priv;

	priv = gtk_graph_view_get_instance_private(GTK_GRAPH_VIEW(obj));

	switch (property_id) {
	case PROP_BG_COLOR:
		data_ptr = g_value_get_boxed(value);
		if (!data_ptr)
			break;
		priv->background_color.red = ((GdkRGBA *)data_ptr)->red;
		priv->background_color.green = ((GdkRGBA *)data_ptr)->green;
		priv->background_color.blue = ((GdkRGBA *)data_ptr)->blue;
		priv->background_color.alpha = ((GdkRGBA *)data_ptr)->alpha;
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, property_id, pspec);
		break;
	}
}

void gtk_graph_view_dispose(GObject *obj)
{
	GtkGraphView *graph_view = GTK_GRAPH_VIEW(obj);
	GtkGraphViewPrivate *priv = gtk_graph_view_get_instance_private(graph_view);

	if (priv->gl_area)
		gtk_gl_area_make_current(GTK_GL_AREA(priv->gl_area));
	g_clear_object(&priv->passthrough_prog);
}

static GParamSpec *gtk_graph_view_properties[N_PROPERTIES] = {NULL};

/**
 * @brief Init class structure of GtkGraphView
 * @param klass Class structure
 */
static void gtk_graph_view_class_init(GtkGraphViewClass *klass)
{
	GObjectClass *oclass;

	oclass = G_OBJECT_CLASS(klass);
	oclass->set_property = gtk_graph_view_set_property;
	oclass->get_property = gtk_graph_view_get_property;
	oclass->dispose = gtk_graph_view_dispose;

	gtk_graph_view_properties[PROP_BG_COLOR] =
			g_param_spec_boxed("background-color", "background color", "Background color of Graph View",
					   GDK_TYPE_RGBA, G_PARAM_READWRITE);

	g_object_class_install_properties(oclass, N_PROPERTIES, gtk_graph_view_properties);
}

static void gl_area_realize(GtkGLArea *area, GtkGraphView *graph_view)
{
	GtkGraphViewPrivate *priv;
	char error_text[512];
	int status;
	const vec3 color = {1.0f, 0.0f, 0.0f};
	const vec3 vertecies[3] = {
		{-0.5f, 0.0f, 0.0f},
		{0.5f, 0.0f, 0.0f},
		{0.0, 0.5f, 0.0f},
	};

	gtk_gl_area_make_current(area);
	priv =	gtk_graph_view_get_instance_private(graph_view);

	priv->passthrough_prog = shimatta_opengl_program_new_from_file(SHADER_DIRECTORY"/passthrough3d.vertex.glsl",
								       NULL,
								       SHADER_DIRECTORY"/monochrome.fragment.glsl");
	status = shimatta_opengl_program_compile(priv->passthrough_prog, error_text, sizeof(error_text));
	if (status) {
		g_warning("Error compiling shader: %s", error_text);
	}

	priv->test = shimatta_opengl_colored_triangle_new(priv->passthrough_prog, color, vertecies);
	shimatta_opengl_graphics_realize(priv->test);
}

static void gl_area_resize(GtkGLArea *area, GtkGraphView *graph_view)
{
	(void)graph_view;

	gtk_gl_area_queue_render(area);
}

/**
 * @brief Init a new GtkGraphView GObject
 * @param obj Object to initialize
 */
static void gtk_graph_view_init(GtkGraphView *obj)
{
	GtkGraphViewPrivate *priv;

	priv = gtk_graph_view_get_instance_private(obj);

	/* Create new GTK GL_Area as only child of the box */
	priv->gl_area = gtk_gl_area_new();
	g_signal_connect(priv->gl_area, "render", G_CALLBACK(gl_render), obj);
	g_signal_connect(priv->gl_area, "realize", G_CALLBACK(gl_area_realize), obj);
	g_signal_connect(priv->gl_area, "resize", G_CALLBACK(gl_area_resize), obj);
	gtk_container_add(GTK_CONTAINER(obj), priv->gl_area);
	gtk_box_set_child_packing(GTK_BOX(obj), priv->gl_area, TRUE, TRUE, 0, GTK_PACK_START);
	gtk_widget_show(priv->gl_area);

	priv->background_color.red = 0.0;
	priv->background_color.green = 0.0;
	priv->background_color.blue = 0.0;
	priv->background_color.alpha = 1.0;
}

GtkWidget *gtk_graph_view_new()
{
	return gtk_widget_new(GTK_TYPE_GRAPH_VIEW, NULL);
}

void gtk_graph_view_set_background_color(GtkGraphView *graph_view, const GdkRGBA *rgba)
{
	GValue color_value = G_VALUE_INIT;

	g_return_if_fail(GTK_IS_GRAPH_VIEW(graph_view));
	g_return_if_fail(rgba != NULL);

	g_value_init(&color_value, GDK_TYPE_RGBA);
	g_value_set_boxed(&color_value, rgba);
	g_object_set_property(G_OBJECT(graph_view), "background-color", &color_value);
	g_value_unset(&color_value);
}
